// ===================================================================
//  Chatty 20/64 - for Commodore 64 and VIC-20 (expanded)
//
// Very crude "proof of concept" for the UDP functions
// of the Wifi 64 Ultra modem.
//
// (c) 2020 by WOZ, Vintage Computing Carinthia
// ===================================================================
//
// Redistribution and use in source and synthezised forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
//
// Redistributions in synthesized form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// Neither the name of the author nor the names of other contributors may
// be used to endorse or promote products derived from this software without
// specific prior written permission.
//
// Commercial use of this code is prohibited.
//
// THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// You are responsible for any legal issues arising from your use of this code.
//
// Please report bugs to the author, but before you do so, please
// make sure that this is not a derivative work and that
// you have the latest version of this file.
// ===================================================================
//
// TODO: make a real program out of that pice of ..... code .... ;-)
//
// At least:
// .) proper character translation
// .) configuration of UDP server
// .) allow configuration and selection of receivers
// .) performance improvements (e.g. see timeout below)
// .) nicer GUI suitable for C64 and VIC-20 (and others)
//
// ===================================================================
//
// USAGE:
//
// Firstly, ensure you use an ULTRA firmware release:
// .) You need firmware "v0.3-06dec2020-woz"
// Secondly, start a terminal program and set up modem:
// .) modem must be set to 300 baud ("at$sb=300")
// .) setup this with a terminal program
// .) WLAN SSID  ("at$ssid=<ssid>")
// .) WLAN Password  ("at$pass=<password>")
// .) own UDP listen port ("atu=<portnum>")
// .) speed dial 7 for UDP receiver ("at&z7=<host>:<portnum>")
// .) store in NVRAM ("at&w")
// Start the program, it is very simple to use (yet):
// .) wait until you see MSG IN and MSG OUT line
// .) start typing (max. 37 char), press enter to send
// .) sent message is copied up, input line is cleared when ready
// .) the '#' disappears when receiving a package
// .) one can not type while data is received

#include <conio.h>
#include <stdlib.h>   // atoi
#include <string.h>  // strcpy

#include "serial.h"

const char debug=0;

#ifdef __VIC20__
#define MSG_LEN 19
#else
#define MSG_LEN 37
#endif

// when we use waitvsync as delay:  PAL: 20ms
//                                 NTSC: 16.7ms
// assume PAL
//
// fix later e.g. with this (reliable PAL/NTSC-detector)
// https://codebase64.org/doku.php?id=base:detect_pal_ntsc
//
void delay (char frames) {
	while (frames) {
		waitvsync();
		frames--;
	}
}

// a bit of math:
//
// time of one byte (8n1) at  300 baud: 33.3ms
//                        at  600 baud: 16.7ms
//                        at 1200 baud:  8.3ms (up to here with kerne method safe)
//                        at 2400 baud:  4.2ms (probably to too fast)
//                        at 4800 baud:  2.1ms (too fast)
//                        at 9600 baud:  1.0ms (UP9600 method - not implemented)
//
// assume 300 baud for now
//
// This timeout stuff works, but would be better to receive based on
// expected data, so it is not required to wait the time out period
// all the times - this makes response even more slower at 300 BAUD!
//
// Thus we also send a BS character, which is ignored by the modem but will
// be send back after executing the command (as echo is on).
//
char send_command(char cmd[], char res[], char wait) {
	char i, timeout;
	unsigned ch;

	// ensure we flush everything we might still have buffered
	set_ser(0);
	// flush_ser();  // unsafe...
	do ch = get_ser();
	while (ch!=0x8000);

	// TX mode, send command
	set_ser(1);
	i=0;
	while (cmd[i]) {
		put_ser(cmd[i++]);
	}
	put_ser(10);
	put_ser(8);

	// switch to RX mode, wait execution
	set_ser(0);

	timeout=50;	// 50x100ms = 5sec timeout
	i=0;
	while(1) {
		ch = get_ser();
		if (ch!=0x8000) {
			char c = ch&0xff;
			// simplify line separator for processing
			//if (c==0x0d) c = ';';
			if (c==0x0a) c = ';';
			if (c==0x08) break;
			if (i==255) break;  // buffer limit
			if (c>=32)
				res[i++] = c;
			timeout=50;
		} else {
			delay(5);  // ~100ms
			timeout--;
			if (!timeout) break;
		}
	}
	res[i]=0;
	return i;
}

// ===================================================================
// Example application follows here...
// ===================================================================

char cmdv[]="atv0";
char cmdi[]="ati";
char cmdz[]="at&z7?";
char cmdu[]="atur7";
char cmdm[]="atum?";
char res[128];

// we need reserves here
char rpos = 0;
char response[255];
// this is limited by given msg. size...
char mpos = 5;
char message[MSG_LEN+5+1]="atum=";

// conio compatible message scrollers
#ifdef __VIC20__
char in_scrn1[MSG_LEN+1]="                   ";
char in_scrn2[MSG_LEN+1]="                   ";
char in_scrn3[MSG_LEN+1]="                   ";
#else
char in_scrn1[MSG_LEN+1]="                                     ";
char in_scrn2[MSG_LEN+1]="                                     ";
char in_scrn3[MSG_LEN+1]="                                     ";
#endif
void msg_in_scroll() {
	strncpy(in_scrn1,in_scrn2,MSG_LEN);
	strncpy(in_scrn2,in_scrn3,MSG_LEN);
	gotoxy(0,11);
	cprintf("> %s",in_scrn1);
	gotoxy(0,12);
	cprintf("> %s",in_scrn2);
	gotoxy(0,13);
#ifdef __VIC20__
	cprintf("                     ");
#else
	cprintf("                                       ");
#endif
}
void msg_in(char *msg) {
	char len=strlen(msg);
	char i;
	strncpy(in_scrn3,msg,MSG_LEN);
	// ASCII -> PETSCII
	for (i=0;i<len;++i) {
		char c=in_scrn3[i];
		if ( ((c>64)&&(c<91)) || ((c>96)&&(c<123)) ) c ^= 32;
		in_scrn3[i]=c;
	}
	while(len<MSG_LEN)in_scrn3[len++]=' ';
	gotoxy(0,13);
	cprintf("> %s",in_scrn3);
}
#ifdef __VIC20__
char out_scrn1[MSG_LEN+1]="                   ";
char out_scrn2[MSG_LEN+1]="                   ";
char out_scrn3[MSG_LEN+1]="                   ";
#else
char out_scrn1[MSG_LEN+1]="                                     ";
char out_scrn2[MSG_LEN+1]="                                     ";
char out_scrn3[MSG_LEN+1]="                                     ";
#endif
void msg_out_scroll() {
	strncpy(out_scrn1,out_scrn2,MSG_LEN);
	strncpy(out_scrn2,out_scrn3,MSG_LEN);
	gotoxy(0,15);
	cprintf("> %s",out_scrn1);
	gotoxy(0,16);
	cprintf("> %s",out_scrn2);
	gotoxy(0,17);
#ifdef __VIC20__
	cprintf("                     ");
#else
	cprintf("                                       ");
#endif
}
void msg_out(char *msg) {
	char len=strlen(msg);
	char i;
	strncpy(out_scrn3,msg,MSG_LEN);
	// ASCII -> PETSCII
	for (i=0;i<len;++i) {
		char c=out_scrn3[i];
		if ( ((c>64)&&(c<91)) || ((c>96)&&(c<123)) ) c ^= 32;
		out_scrn3[i]=c;
	}
	while(len<MSG_LEN)out_scrn3[len++]=' ';
	gotoxy(0,17);
	cprintf("# %s",out_scrn3);
}

// ===============================
// === MAIN                    ===
// ===============================

void main(void) {
	char sx, sy;
	char i,len;
	char err=0;
	
	screensize(&sx,&sy);

	clrscr();
	cprintf("Chatty v0.3beta\r\n");
	cprintf("(c) 2020 WOZ/vcc\r\n");
	cprintf("Screen: %d x %d", sx, sy);
	gotoxy(0,3);
	sy--;

	open_ser();

	// machine-communication mode
	send_command(cmdv,res,20);

	// show actual network settings
	len = send_command(cmdi,res,20);

	// find UTP port
	// result is something like:
	// "ati;1,0,0,6798,0;0;"
	gotoxy(0,3);
	cprintf("RX:");
	if (len<12) {
		cprintf("ERROR ");
		if (debug) cprintf("%d: %s\n\r", len, res);
		err=1;
	} else {
		char pos1=0, pos2=0;
		if (debug) cprintf("%d: %s\n\r", len, res);
		for(i=len;i>0;i--) {
			if (res[i]==',')
				if (pos2) { pos1=i; break; }
				else pos2=i;
		}
		if (debug) cprintf("%d - %d, ", pos1, pos2);
		res[pos2]=0; ++pos1;
		if (pos2>pos1) {
			cprintf("%s", res+pos1);
		} else {
			cprintf("OFF");
			err=1;
		}
	}

	if (err) goto fail;
	
	// show last speed dial (used to chat)
	len = send_command(cmdz,res,200);

	// find recipient IP and port
	// result looks like:
	// "at&z9?;;192.168.1.12:123;0;"
	gotoxy(0,5);
	cprintf("TX:");
	if (len<7) {
		cprintf("ERROR ");
		if (debug) cprintf("%d: %s\n\r", len, res);
		err=1;
	} else {
		char pos1=0, pos2=0;
		if (debug) cprintf("%d: %s, ", len, res);
		for(i=len-2;i>0;i--) {
			if (res[i]==';')
				if (pos2) { pos1=i; break; }
				else pos2=i;
		}
		if (debug) cprintf("%d - %d, ", pos1, pos2);
		res[pos2]=0; ++pos1;
		if (pos2>pos1) {
			cprintf("%s", res+pos1);
		} else {
			cprintf("-");
			err=1;
		}
	}

	if (err) goto fail;
	
	// set last speed dial as receiver
	len = send_command(cmdu,res,200);

	// find recipient IP and port
	// result looks like:
	// "atur7;0;"
	gotoxy(0,7);
	cprintf("Use speed dial:");
	if (len<7) {
		cprintf("ERROR ");
		if (debug) cprintf("%d: %s\n\r", len, res);
		err=1;
	} else {
		char pos1=0;
		if (debug) cprintf("%d: %s, ", len, res);
		for(i=len-2;i>0;i--) {
			if (res[i]=='r') {
				pos1=i; break;
			}
		}
		if (debug) cprintf("%d, ", pos1);
		if (pos1) {
			res[i+2]=0;
			cprintf("%s", res+pos1+1);
		} else {
			cprintf("-");
			err=1;
		}
	}

	if (err) goto fail;
	
	gotoxy(0,10);
#ifdef __VIC20__
	cprintf("=MSG IN===============");
#else
	cprintf("=MSG IN=================================");
#endif
	msg_in_scroll();
	msg_in("");
	gotoxy(0,14);
#ifdef __VIC20__
	cprintf("=MSG OUT==============");
#else
	cprintf("=MSG OUT================================");
#endif
	msg_out_scroll();
	msg_out("");
	gotoxy(0,18);
#ifdef __VIC20__
	cprintf("======================");
#else
	cprintf("========================================");
#endif

	while (1) {
        int ch = get_ser();
        if (ch!=0x8000) {
            char c = ch&0xff;
			if (c=='*') {
				char pos1=0, pos2=0;
				msg_in_scroll();
				// we got an asterics, get message. looks like:
				// ";atum?;192.168.1.5:6788;7:GOTCHA_;0;"
				len = send_command(cmdm,res,200);
				if (debug) cprintf("%d: %s\n\r", len, res);
				for(i=len-2;i>0;i--) {
					if (res[i]==';') {
						pos2=i;
					}
					if (res[i]==':') {
						pos1=i;
						break;
					}
				}
				if (debug) cprintf("%d - %d, ", pos1, pos2);
				res[pos2]=0; ++pos1;
				if (pos2>pos1) {
					strncpy(response,res+pos1,MSG_LEN);
					msg_in(response);
#ifdef __VIC20__
					// vic has smaller screen, so we may need 2 lines...
					if ((pos2-pos1)>MSG_LEN) {
						msg_in_scroll();
						strncpy(response,res+pos1+MSG_LEN,MSG_LEN);
						msg_in(response);
					}
#endif
				} else {
					msg_in("-");
				}
			}
		} else
        if (kbhit()) {
            char c = cgetc();
			if (debug) gotoxy(27,21);
			if (debug) cprintf("%03d-0x%02x",c,c);
			// PETSCII --> ASCII
            if (c>128) c ^= 160;
			if ( ((c>64)&&(c<91)) || ((c>96)&&(c<123)) ) c ^= 32;
			if (debug) gotoxy(27,22);
			if (debug) cprintf("%03d-0x%02x",c,c);			
            if ((c>=32) && (mpos<(MSG_LEN+5))) {
				message[mpos++]=c;
				message[mpos]=0;
			} else {
				if ((c==20) && (mpos>5))
					message[--mpos]=0;
				else if ((c==13) && (mpos>5)) {
					msg_out_scroll();
					len = send_command(message,res,200);
					if (debug) cprintf("%d: %s\n\r", len, res);
					mpos=5;
					message[mpos]=0;
					msg_out(message+5);
				} else
				if (debug) cprintf("%d", c);
			}
			msg_out(message+5);
        };
	}

// year, I know - not really structured programming here...
fail:
	close_ser();
	gotoxy(0,21);
	cprintf("Good bye!");
	return;
}
